/*-
 * Copyright (c) 2004 Max Khon
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <sys/param.h>
#include <sys/bio.h>
#include <sys/endian.h>
#include <sys/errno.h>
#include <sys/kernel.h>
#include <sys/lock.h>
#include <sys/mutex.h>
#include <sys/malloc.h>
#include <sys/systm.h>

#include <geom/geom.h>
#include <net/zlib.h>

MALLOC_DEFINE(M_GEOM_UGZ, "GEOM UGZ", "GEOM UGZ data structures");

#define UGZ_CLASS_NAME	"UGZ"

/*
 * Maximum allowed valid block size (to prevent foot-shooting)
 */
#define MAX_BLKSZ	(MAXPHYS - MAXPHYS / 1000 - 12)

/*
 * Integer values (block size, number of blocks, offsets)
 * are stored in big-endian (network) order on disk and struct cloop_header
 * and in native order in struct g_ugz_softc
 */

#define CLOOP_MAGIC_LEN 128
static char CLOOP_MAGIC_START[] = "#!/bin/sh\n";

struct cloop_header {
	char magic[CLOOP_MAGIC_LEN];	/* cloop magic */
	uint32_t blksz;			/* block size */
	uint32_t nblocks;		/* number of blocks */
};

#undef GEOM_UGZ_DEBUG
#ifdef GEOM_UGZ_DEBUG
#define DPRINTF(a)	printf a
#else
#define DPRINTF(a)
#endif

struct g_ugz_softc {
	uint32_t blksz;			/* block size */
	uint32_t nblocks;		/* number of blocks */
	uint64_t *offsets;
};

static void
g_ugz_softc_free(struct g_ugz_softc *sc)
{
	if (sc->offsets != NULL)
		free(sc->offsets, M_GEOM_UGZ);
	free(sc, M_GEOM_UGZ);
}

static void *
z_alloc(void *nil, uint type, uint size)
{
	void *ptr;

	ptr = malloc(type * size, M_GEOM_UGZ, M_NOWAIT);
	return ptr;
}

static void
z_free(void *nil, void *ptr)
{
	free(ptr, M_GEOM_UGZ);
}

static void
g_ugz_done(struct bio *bp)
{
	int err;
	struct bio *bp2;
	z_stream zs;
	struct g_provider *pp, *pp2;
	struct g_consumer *cp;
	struct g_geom *gp;
	struct g_ugz_softc *sc;
	off_t pos, upos, uoff;
	char *buf;
	uint32_t start_blk, i;
	size_t bsize;

	bp2 = bp->bio_parent;
	pp = bp2->bio_to;
	gp = pp->geom;
	cp = LIST_FIRST(&gp->consumer);
	pp2 = cp->provider;
	sc = gp->softc;
	buf = bp->bio_data;
	buf -= sc->blksz;
	DPRINTF(("%s: done\n", gp->name));

	bp2->bio_error = bp->bio_error;
	if (bp2->bio_error != 0)
		goto done;

	/*
	 * Uncompress data.
	 */
	zs.zalloc = z_alloc;
	zs.zfree = z_free;
	err = inflateInit(&zs);
	if (err != Z_OK) {
		bp2->bio_error = EIO;
		goto done;
	}
	start_blk = bp2->bio_offset / sc->blksz;
	bsize = pp2->sectorsize;
	pos = sc->offsets[start_blk] % bsize;
	upos = 0;
	uoff = bp2->bio_offset % sc->blksz;
	DPRINTF(("%s: done: start_blk %d, pos %lld, upos %lld, uoff %lld (%lld, %d, %d)\n",
	    pp->geom->name, start_blk, pos, upos, uoff,
	    bp2->bio_offset, sc->blksz, bsize));
	for (i = start_blk; upos < bp2->bio_length; i++) {
		off_t len, ulen;

		len = sc->offsets[i + 1] - sc->offsets[i];
		zs.next_in = bp->bio_data + pos;
		zs.avail_in = len;
		zs.next_out = buf;
		zs.avail_out = sc->blksz;

		err = inflate(&zs, Z_FINISH);
		if (err != Z_STREAM_END) {
			inflateEnd(&zs);
			bp2->bio_error = EIO;
			goto done;
		}
		ulen = MIN(sc->blksz, bp2->bio_length - upos);
		DPRINTF(("%s: done: inflated %lld + %lld -> %lld + %lld + %lld\n",
		    pp->geom->name,
		    pos, len,
		    uoff, upos, ulen));

		memcpy(bp2->bio_data + upos, buf + uoff, ulen);
		pos += len;
		upos += ulen;
		bp2->bio_completed += ulen;
		err = inflateReset(&zs);
		if (err != Z_OK) {
			inflateEnd(&zs);
			bp2->bio_error = EIO;
			goto done;
		}
	}
	err = inflateEnd(&zs);
	if (err != Z_OK) {
		bp2->bio_error = EIO;
		goto done;
	}

done:
	/*
	 * Finish processing the request.
	 */
	DPRINTF(("%s: done: (%d, %lld, %ld)\n",
	    pp->geom->name, bp2->bio_error, bp2->bio_completed, bp2->bio_resid));
	free(buf, M_GEOM_UGZ);
	g_destroy_bio(bp);
	g_io_deliver(bp2, bp2->bio_error);
}

static void
g_ugz_start(struct bio *bp)
{
	struct bio *bp2;
	struct g_provider *pp, *pp2;
	struct g_geom *gp;
	struct g_consumer *cp;
	struct g_ugz_softc *sc;
	uint32_t start_blk, end_blk;
	char *buf;
	size_t bsize;

	pp = bp->bio_to;
	gp = pp->geom;
	DPRINTF(("%s: start (%d)\n", gp->name, bp->bio_cmd));

	if (bp->bio_cmd != BIO_READ) {
		g_io_deliver(bp, EOPNOTSUPP);
		return;
	}

	cp = LIST_FIRST(&gp->consumer);
	pp2 = cp->provider;
	sc = gp->softc;

	bp2 = g_clone_bio(bp);
	if (bp2 == NULL) {
		g_io_deliver(bp, ENOMEM);
		return;
	}
	bp2->bio_done = g_ugz_done;

	start_blk = bp->bio_offset / sc->blksz;
	end_blk = start_blk + (bp->bio_length + sc->blksz - 1) / sc->blksz;
	bsize = pp2->sectorsize;
	KASSERT(start_blk <= sc->nblocks,
		("start_blk out of range"));
	KASSERT(end_blk < sc->nblocks,
		("end_blk out of range"));
	DPRINTF(("%s: start (%d..%d), %s: %d + %lld, %s: %d + %lld\n",
	    gp->name, start_blk, end_blk,
	    pp->name, pp->sectorsize, pp->mediasize,
	    pp2->name, pp2->sectorsize, pp2->mediasize));
	bp2->bio_offset = sc->offsets[start_blk] - sc->offsets[start_blk] % bsize;
	bp2->bio_length = sc->offsets[end_blk] - bp2->bio_offset;
	bp2->bio_length = (bp2->bio_length + bsize - 1) / bsize * bsize;
	DPRINTF(("%s: start %lld + %lld -> %lld + %lld -> %lld + %lld\n",
	    gp->name,
	    bp->bio_offset, bp->bio_length,
	    sc->offsets[start_blk], sc->offsets[end_blk] - sc->offsets[start_blk],
	    bp2->bio_offset, bp2->bio_length));
	buf = malloc(sc->blksz + bp2->bio_length, M_GEOM_UGZ, M_NOWAIT);
	if (buf == NULL) {
		g_io_deliver(bp, ENOMEM);
		return;
	}
	bp2->bio_data = buf + sc->blksz;

	g_io_request(bp2, cp);
	DPRINTF(("%s: start ok\n", gp->name));
}

static void
g_ugz_orphan(struct g_consumer *cp)
{
	struct g_geom *gp;

	g_trace(G_T_TOPOLOGY, "g_ugz_orphan(%p/%s)", cp, cp->provider->name);
	g_topology_assert();
	KASSERT(cp->provider->error != 0,
		("g_ugz_orphan with error == 0"));

	gp = cp->geom;
	g_ugz_softc_free(gp->softc);
	g_wither_geom(gp, cp->provider->error);
}

static int
g_ugz_access(struct g_provider *pp, int dr, int dw, int de)
{
	struct g_geom *gp;
	struct g_consumer *cp;

	gp = pp->geom;
	cp = LIST_FIRST(&gp->consumer);
	KASSERT (cp != NULL, ("g_ugz_access but no consumer"));

	/* On first open, grab an extra "exclusive" bit */
	if (cp->acr == 0 && cp->acw == 0 && cp->ace == 0)
		de++;
	/* ... and let go of it on last close */
	if ((cp->acr + dr) == 0 && (cp->acw + dw) == 0 && (cp->ace + de) == 1)
		de--;
	return (g_access_rel(cp, dr, dw, de));
}

static void
g_ugz_spoiled(struct g_consumer *cp)
{
	struct g_geom *gp;

	gp = cp->geom;
	g_trace(G_T_TOPOLOGY, "g_ugz_spoiled(%p/%s)", cp, gp->name);
	g_topology_assert();

	g_ugz_softc_free(gp->softc);
	g_wither_geom(gp, ENXIO);
}

static struct g_geom *
g_ugz_taste(struct g_class *mp, struct g_provider *pp, int flags)
{
	int error;
	uint32_t i, total_offsets, offsets_read, blk;
	void *buf;
	struct cloop_header *header;
	struct g_consumer *cp;
	struct g_geom *gp;
	struct g_provider *pp2;
	struct g_ugz_softc *sc;

	g_trace(G_T_TOPOLOGY, "vol_taste(%s,%s)", mp->name, pp->name);
	g_topology_assert();
	buf = NULL;

	/*
	 * Create geom instance.
	 */
	gp = g_new_geomf(mp, "%s.ugz", pp->name);
	gp->start = g_ugz_start;
	gp->orphan = g_ugz_orphan;
	gp->access = g_ugz_access;
	gp->spoiled = g_ugz_spoiled;
	cp = g_new_consumer(gp);
	error = g_attach(cp, pp);
	if (error == 0)
		error = g_access_rel(cp, 1, 0, 0);
	if (error) {
		g_wither_geom(gp, ENXIO);
		return (NULL);
	}
	g_topology_unlock();

	/*
	 * Read cloop header, look for CLOOP magic, perform
	 * other validity checks.
	 */
	DPRINTF(("%s: media sectorsize %u, mediasize %lld\n",
	    gp->name, pp->sectorsize, pp->mediasize));
	buf = g_read_data(cp, 0, pp->sectorsize, &error);
	if (buf == NULL || error != 0)
		goto err;
	header = (struct cloop_header *) buf;
	if (strncmp(header->magic, CLOOP_MAGIC_START,
		    sizeof(CLOOP_MAGIC_START) - 1) != 0) {
		DPRINTF(("%s: no CLOOP magic\n", gp->name));
		goto err;
	}
	if (header->magic[0x0b] != 'V' || header->magic[0x0c] < '2') {
		DPRINTF(("%s: image version too old\n", gp->name));
		goto err;
	}

	/*
	 * Initialize softc and read offsets.
	 */
	sc = malloc(sizeof(*sc), M_GEOM_UGZ, M_WAITOK);
	gp->softc = sc;
	sc->blksz = ntohl(header->blksz);
	sc->nblocks = ntohl(header->nblocks);
	if (sc->blksz % 512 != 0) {
		printf("%s: block size (%u) should be multiple of 512.\n",
		    gp->name, sc->blksz);
		goto err;
	}
	if (sc->blksz > MAX_BLKSZ) {
		printf("%s: block size (%u) should not be larger than %d.\n",
		    gp->name, sc->blksz, MAX_BLKSZ);
	}
	total_offsets = sc->nblocks + 1;
	if (sizeof(struct cloop_header) +
	    total_offsets * sizeof(uint64_t) > pp->mediasize) {
		printf("%s: media too small for %u blocks\n",
		       gp->name, sc->nblocks);
		goto err;
	}
	sc->offsets = malloc(
	    total_offsets * sizeof(uint64_t), M_GEOM_UGZ, M_WAITOK);
	offsets_read = MIN(total_offsets,
	    (pp->sectorsize - sizeof(*header)) / sizeof(uint64_t));
	for (i = 0; i < offsets_read; i++)
		sc->offsets[i] = be64toh(((uint64_t *) (header + 1))[i]);
	DPRINTF(("%s: %u offsets in the first sector\n",
	       gp->name, offsets_read));
	for (blk = 1; offsets_read < total_offsets; blk++) {
		uint32_t nread;

		free(buf, M_GEOM_UGZ);
		buf = g_read_data(
		    cp, blk * pp->sectorsize, pp->sectorsize, &error);
		if (buf == NULL || error != 0)
			goto err;
		nread = MIN(total_offsets - offsets_read,
		     pp->sectorsize / sizeof(uint64_t));
		DPRINTF(("%s: %u offsets read from sector %d\n",
		    gp->name, nread, blk));
		for (i = 0; i < nread; i++) {
			sc->offsets[offsets_read + i] =
			    be64toh(((uint64_t *) buf)[i]);
		}
		offsets_read += nread;
	}
	DPRINTF(("%s: done reading offsets\n", gp->name));

	g_topology_lock();
	pp2 = g_new_providerf(gp, "%s", gp->name);
	pp2->sectorsize = 512;
	pp2->mediasize = sc->nblocks * sc->blksz;
        pp2->flags = pp->flags & G_PF_CANDELETE;
        if (pp->stripesize > 0) {
                pp2->stripesize = pp->stripesize;
                pp2->stripeoffset = pp->stripeoffset;
        }
	g_error_provider(pp2, 0);
	g_access_rel(cp, -1, 0, 0);

	DPRINTF(("%s: taste ok (%d, %lld), (%d, %d), %x\n",
	    gp->name,
	    pp2->sectorsize, pp2->mediasize,
	    pp2->stripeoffset, pp2->stripesize, pp2->flags));
	printf("%s: %u x %u blocks\n",
	       gp->name, sc->nblocks, sc->blksz);
	return (gp);

err:
	g_topology_lock();
	g_access_rel(cp, -1, 0, 0);
	if (buf != NULL)
		free(buf, M_GEOM_UGZ);
	if (gp->softc != NULL)
		g_ugz_softc_free(gp->softc);
	g_wither_geom(gp, ENXIO);
	return (NULL);
}

static int
g_ugz_destroy_geom(struct gctl_req *req, struct g_class *mp, struct g_geom *gp)
{
	g_trace(G_T_TOPOLOGY, "g_ugz_destroy_geom(%s, %s)", mp->name, gp->name);
	g_topology_assert();

	g_ugz_spoiled(LIST_FIRST(&gp->consumer));
	return (0);
}

static struct g_class g_ugz_class = {
	.name = UGZ_CLASS_NAME,
	.taste = g_ugz_taste,
	.destroy_geom = g_ugz_destroy_geom,
};

DECLARE_GEOM_CLASS(g_ugz_class, g_ugz);
